// ===========Oooo==========================================Oooo========
// =  Copyright (C) 2014 Shanghai Gowin Semiconductor Technology Co.,Ltd.
// =                     All rights reserved.
// =====================================================================
//
//  __      __      __
//  \ \    /  \    / /   [File name   ] extractor.v
//   \ \  / /\ \  / /    [Description ] comb-video extractor
//    \ \/ /  \ \/ /     [Timestamp   ] 2022/08/04
//     \  /    \  /      [version     ] 1.0
//      \/      \/       
// --------------------------------------------------------------------
// Code Revision History :
// --------------------------------------------------------------------
// Ver: | Author        | Mod. Date  | Changes Made:
// V1.0 | Huang Mingtao | 2022/08/04 | Initial version
// ===========Oooo==========================================Oooo========

module DPI_EXTRACTOR #
(   parameter [ 2 : 0]  InPolarity      = 3'b001    //! MSB - VS -> HS -> DE - LSB
,   parameter [ 2 : 0]  OutPolarity     = 3'b001    //! MSB - VS -> HS -> DE - LSB
,   parameter [31 : 0]  ExtrPattern     = 32'b11
,   parameter           ExtrPatternLen  = 2         //! 2~32
,   parameter           ExtrStartX      = 0
,   parameter           OutputWidth     = 180
,   parameter           OutputHFP       = 40
,   parameter           OutputHSW       = 40
,   parameter           OutputHBP       = 40
,   parameter           OutputHeight    = 240
,   parameter           OutputVFP       = 4
,   parameter           OutputVSW       = 2
,   parameter           OutputVBP       = 4
,   parameter           VSyncDelay      = 2000
)
(   input           I_RSTN

,   input           I_CLK_IN
,   input           I_VS_IN
,   input           I_HS_IN
,   input           I_DE_IN
,   input  [23:0]   I_RGB_IN
//!
,   input           I_CLK_OUT
,   output          O_VS
,   output          O_HS
,   output          O_DE
,   output [23:0]   O_RGB
);

localparam  OutVSPolarity   = OutPolarity[2]        ;
localparam  OutHSPolarity   = OutPolarity[1]        ;
localparam  OutDEPolarity   = OutPolarity[0]        ;

localparam  RealPatternLen  = ExtrPatternLen - 1    ;

localparam  AlignedWidth    = 2**$clog2(OutputWidth);
localparam  MemDepth        = AlignedWidth*4        ;
localparam  MemAddr         = $clog2(MemDepth)      ;

localparam  DlyCntInit      = VSyncDelay + 18       ;
localparam  DlyCntBits      = $clog2(DlyCntInit)    ;

localparam  ExtrEndX        = ExtrStartX + OutputWidth;
localparam  OutputHTT       = OutputWidth + OutputHFP + OutputHSW + OutputHBP;
localparam  OutputVTT       = OutputHeight + OutputVFP + OutputVSW + OutputVBP;

reg                 ivs_r           ;
reg                 ihs_r           ;
reg                 ide_r           ;
reg                 ivs_begin_r     ;
reg                 ihs_begin_r     ;
reg                 iworking        ;

reg  [DlyCntBits-1:0] ivs_pulse_cnt ;
reg                 ivs_pulse_gen   ;
reg  [15:0]         pixel_cnt       ;
reg                 iact_line       ;
reg  [ 4:0]         pattern_cnt     ;
reg  [31:0]         pattern         ;
reg  [9:0]          ivcnt           ;

reg                 buf_wr          ;
reg  [MemAddr-1:0]  buf_wa          ;
reg  [23:0]         buf_wd          ;


wire ivs_w = (I_VS_IN == InPolarity[2]);
wire ihs_w = (I_HS_IN == InPolarity[1]);
wire ide_w = (I_DE_IN == InPolarity[0]);

wire ivs_begin_w =  ivs_w & ~ivs_r;
wire ivs_end_w   = ~ivs_w &  ivs_r;
wire ihs_begin_w =  ihs_w & ~ihs_r;
wire ihs_end_w   = ~ihs_w &  ihs_r;
wire ide_begin_w =  ide_w & ~ide_r;
wire ide_end_w   = ~ide_w &  ide_r;

always @(negedge I_RSTN or posedge I_CLK_IN)
if (!I_RSTN)    {ivs_r, ihs_r, ide_r}  <= 0;
else            {ivs_r, ihs_r, ide_r}  <= {ivs_w, ihs_w, ide_w};

always @(negedge I_RSTN or posedge I_CLK_IN)
if (!I_RSTN)    {ivs_begin_r, ihs_begin_r} <= 2'b00;
else            {ivs_begin_r, ihs_begin_r} <= {ivs_begin_w, ihs_begin_w};

always @(negedge I_RSTN or posedge I_CLK_IN)
if (!I_RSTN)        iworking <= 1'b0;
else if (!iworking) iworking <= ivs_begin_w;

always @(posedge I_CLK_IN)
if (!iworking)              ivs_pulse_cnt <= 0;
else if (ivs_begin_r)       ivs_pulse_cnt <= DlyCntInit;
else if (ivs_pulse_cnt>0)   ivs_pulse_cnt <= ivs_pulse_cnt - 1'b1;

always @(negedge I_RSTN or posedge I_CLK_IN)
if (!I_RSTN | !iworking)    ivs_pulse_gen <= 1'b0;
else                        ivs_pulse_gen <= (ivs_pulse_cnt > 0) & (ivs_pulse_cnt <= 16);

always @(posedge I_CLK_IN)
if (!iworking | ihs_w)      pixel_cnt <= 'd0;
else if (ide_w)             pixel_cnt <= pixel_cnt + 2'd1;

always @(posedge I_CLK_IN)
if (!iworking)  iact_line <= 1'b0;
else            iact_line <= iact_line ? !ihs_begin_w : ide_begin_w;

always @(posedge I_CLK_IN)
if (!iworking | ivs_w)          pattern_cnt <= 5'd0;
else if (iact_line&ihs_begin_w) pattern_cnt <= (pattern_cnt >= RealPatternLen) ? 5'd0 : (pattern_cnt + 1'b1);

always @(posedge I_CLK_IN)
if (!iworking | ivs_w)          pattern <= ExtrPattern;
else if (iact_line&ihs_begin_w) pattern <= (pattern_cnt >= RealPatternLen) ? ExtrPattern : {1'b0, pattern[31:1]};

always @(negedge I_RSTN or posedge I_CLK_IN)
if (!I_RSTN)                            buf_wr  <= 1'b0;
else if (!iworking|!ide_w|!pattern[0])  buf_wr  <= 1'b0;
else                                    buf_wr  <= (pixel_cnt >= ExtrStartX) & (pixel_cnt < ExtrEndX);

always @(posedge I_CLK_IN)
if (!iworking | ihs_w)  buf_wa[MemAddr-3:0] <= 'd0;
else if (buf_wr)        buf_wa[MemAddr-3:0] <= buf_wa[MemAddr-3:0] + 1'b1;

always @(posedge I_CLK_IN)
if (!iworking | ivs_w)                      buf_wa[MemAddr-1-:2] <= 'd0;
else if (iact_line&ihs_begin_w&pattern[0])  buf_wa[MemAddr-1-:2] <= buf_wa[MemAddr-1-:2] + 1'b1;

always @(posedge I_CLK_IN)
if (!iworking | ivs_w)                      ivcnt <= 'd0;
else if (iact_line&ihs_begin_w&pattern[0])  ivcnt <= ivcnt + 1'b1;

always @(posedge I_CLK_IN)
if (!iworking)      buf_wd  <= 'b0;
else if (ide_w)     buf_wd  <= I_RGB_IN;

//! ********* DPRAM for saving pixel *********
//!
reg  [23:0]         mem_data    [MemDepth-1:0] /* synthesis syn_ramstyle="block_ram" */;
reg  [23:0]         mem_rqout_r /* synthesis syn_preserve = 1 */;
//!
wire                mem_wren    /* synthesis syn_preserve = 1 */;
wire [MemAddr-1:0]  mem_waddr   /* synthesis syn_preserve = 1 */;
wire [23:0]         mem_wdin    /* synthesis syn_preserve = 1 */;
//!
wire                mem_rden    /* synthesis syn_preserve = 1 */;
wire [MemAddr-1:0]  mem_raddr   /* synthesis syn_preserve = 1 */;

assign mem_wren  = buf_wr   ;
assign mem_waddr = buf_wa   ;
assign mem_wdin  = buf_wd   ;

always @(posedge I_CLK_IN)
if (mem_wren)       mem_data[mem_waddr] <= mem_wdin;

always @(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)        mem_rqout_r <= 0;
else if (mem_rden)  mem_rqout_r <= mem_data[mem_raddr];
//
//! ********* END OF DPRAM *********

reg                 oworking    ;
reg  [1:0]          ivs_o_sync  ;
reg                 ovs         ;
reg                 ohs         ;
reg                 ode         ;
reg                 ovs_trigger ;

reg                 oact_line   ;
reg                 oinc_line   ;
reg  [15:0]         ohcnt       ;
reg  [15:0]         ovcnt       ;
reg                 o_vs_reg    ;
reg                 o_hs_reg    ;
reg                 o_de_reg    ;
reg  [23:0]         o_rgb_reg   ;

reg                 buf_rd      ;
reg  [MemAddr-1:0]  buf_ra      ;

wire [15:0]         ohcnt_inc   = ohcnt + 1'b1 ;
wire [15:0]         ovcnt_inc   = ovcnt + 1'b1 ;

assign mem_rden     = buf_rd    ;
assign mem_raddr    = buf_ra    ;

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)        ivs_o_sync  <= 2'b00;
else                ivs_o_sync  <= {ivs_o_sync[0], ivs_pulse_gen};

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)        oworking    <= 1'b0;
else if (oworking)  oworking    <= iworking;
else                oworking    <= ivs_o_sync[0] & ~ivs_o_sync[1];

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)        ovs_trigger <= 1'b0;
else                ovs_trigger <= ivs_o_sync[0] & ~ivs_o_sync[1];

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)                    ohcnt  <= 16'd0;
else if (!oworking|ovs_trigger) ohcnt  <= 16'd0;
else                            ohcnt  <= (ohcnt_inc < OutputHTT) ? ohcnt_inc : 16'd0;

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)        oinc_line <= 1'b0;
else if (!oworking) oinc_line <= 1'b0;
else                oinc_line <= (ohcnt_inc == OutputHTT);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)                    ovcnt   <= 16'd0;
else if (!oworking|ovs_trigger) ovcnt   <= 16'd0;
else if (oinc_line)             ovcnt   <= ovcnt_inc;

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)                    ovs     <= 1'b0;
else if (!oworking)             ovs     <= 1'b0;
else                            ovs     <= (ovcnt > 16'd0) & (ovcnt <= OutputVSW);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)                                ohs <= 1'b0;
else if (!oworking|(ovcnt_inc> OutputVTT))  ohs <= 1'b0;
else                                        ohs <= (ohcnt_inc >  OutputHFP) 
                                                 & (ohcnt_inc <= OutputHFP + OutputHSW);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)            oact_line <= 1'b0;
else if (!oworking)     oact_line <= 1'b0;
else                    oact_line <= (ovcnt >  OutputVSW + OutputVBP) 
                                   & (ovcnt <= OutputVSW + OutputVBP + OutputHeight);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)                    buf_rd  <= 1'b0;
else if (!oworking|!oact_line)  buf_rd  <= 1'b0;
else                            buf_rd  <= (ohcnt_inc <  OutputHTT)
                                         & (ohcnt_inc >= OutputHFP + OutputHSW + OutputHBP);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN) ode <= 1'b0;
else         ode <= buf_rd;

always @(posedge I_CLK_OUT)
if (!oworking | ohs)    buf_ra[MemAddr-3:0] <= 'd0;
else if (buf_rd)        buf_ra[MemAddr-3:0] <= buf_ra[MemAddr-3:0] + 1'b1;

always @(posedge I_CLK_OUT)
if (!oworking | ovs)    buf_ra[MemAddr-1-:2]  <= 'd0;
else if (~buf_rd & ode) buf_ra[MemAddr-1-:2]  <= buf_ra[MemAddr-1-:2] + 1'b1;

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)            o_vs_reg <= !OutVSPolarity;
else                    o_vs_reg <= (ovs == OutVSPolarity);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)            o_hs_reg <= !OutHSPolarity;
else                    o_hs_reg <= (ohs == OutHSPolarity);

always@(negedge I_RSTN or posedge I_CLK_OUT)
if (!I_RSTN)            o_de_reg <= !OutDEPolarity;
else                    o_de_reg <= (ode == OutDEPolarity);

always@(posedge I_CLK_OUT)
o_rgb_reg <= ode ? mem_rqout_r : 0;

assign O_VS     = o_vs_reg  ;
assign O_HS     = o_hs_reg  ;
assign O_DE     = o_de_reg  ;
assign O_RGB    = o_rgb_reg ;

endmodule
